import paramiko
import os
from pathlib import Path
# from io import StringIO


def gen_keypair(ssh_folder=None, password=None):
    """Create a set of public and private keys and save them in the given
    folder.

    Parameters
    ----------
    ssh_folder : str
        folder where the keys are saved.
    password : str
        encrypt the private key with a password.
    """

    if not ssh_folder:
        ssh_folder = os.path.join(Path.home(), '.ssh')
    key = paramiko.RSAKey.generate(4096)
    with open(os.path.join(ssh_folder, "id_rsa.pub"), "w") as key_file:
        key_file.write("ssh-rsa {}".format(key.get_base64()))
    with open(os.path.join(ssh_folder, "id_rsa"), "w") as key_file:
        key.write_private_key(key_file, password)

    return "Key pair successfully generated in {}".format(ssh_folder)


def add_pubkey_to_server(bot, ssh_folder, os_type='linux'):
    """Adds the public key to the server's list.

    Parameters
    ----------
    bot : obj
        `sshBot` object to access the server.
    ssh_folder : str
        folder where the keys are stored.

    Warnings
    --------
    This works only on linux servers.
    """
    with open(os.path.join(ssh_folder, 'id_rsa.pub'), "r") as pubkey_file:
        pubkey = pubkey_file.readline()
    bot.execute_cmds(commands=['(umask 077 && test -d ~/.ssh || mkdir ~/.ssh)',
                               '(umask 077 && touch ~/.ssh/authorized_keys)',
                               'echo {} >>  ~/.ssh/authorized_keys'.format(pubkey)],
                     close_connection=False,
                     verbose=False)
    return ("public key successfully added. Try to run `ssh hostname`")

    # def connect_with_pvtkey():
    #     k = paramiko.RSAKey.from_private_key_file("/home/fr/.ssh/id_rsa", password="Sxcdews23")
    #     c = paramiko.SSHClient()
    #     c.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    #     print("connecting")
    #     c.connect(hostname="192.168.0.199", username="franaudo", pkey=k)
    #     print("connected")
    #     # commands = [ "/home/ubuntu/firstscript.sh", "/home/ubuntu/secondscript.sh" ]
    #     # for command in commands:
    #     # 	print "Executing {}".format( command )
    #     # 	stdin , stdout, stderr = c.exec_command(command)
    #     # 	print stdout.read()
    #     # 	print( "Errors")
    #     # 	print stderr.read()
    #     # c.close()

    # # def ssh_key_gen(length=2048, type='rsa', password=None, username='jumpserver', hostname=None):
    # #     """Generate user ssh private and public key

    # #     Use paramiko RSAKey generate it.
    # #     :return private key str and public key str
    # #     """

    # #     if hostname is None:
    # #         hostname = os.uname()[1]

    # #     f = StringIO()
    # #     try:
    # #         if type == 'rsa':
    # #             private_key_obj = paramiko.RSAKey.generate(length)
    # #         elif type == 'dsa':
    # #             private_key_obj = paramiko.DSSKey.generate(length)
    # #         else:
    # #             raise IOError('SSH private key must be `rsa` or `dsa`')
    # #         private_key_obj.write_private_key(f, password=password)
    # #         private_key = f.getvalue()
    # #         public_key = ssh_pubkey_gen(private_key_obj, username=username, hostname=hostname)
    # #         return private_key, public_key
    # #     except IOError:
    # #         raise IOError('These is error when generate ssh key.')

    # def create_keypair():

    #     key = paramiko.RSAKey.generate(4096)
    #     print(key.get_base64())  # print public key
    #     key.write_private_key(sys.stdout)  # print private key

    #     with open("/home/fr/.ssh/pubkey", "w") as key_file:
    #         key_file.write(key.get_base64())

    #     key.write_private_key("/home/fr/.ssh/pvtkey")


if __name__ == "__main__":
    from codebots.bots import sshBot
    bot = sshBot(config_file=None, hostname="192.168.0.199", username="franaudo", password="Sxcdews23", pvtkey="")
    add_pubkey_to_server(bot, path_local="C:/Users/franaudo/.ssh/pubkey")



import requests
import os


from ._bot import BaseBot

__all__ = [
    'TeleBot'
]


class TeleBot(BaseBot):
    """Bot that automatically sends messages over telegram

    Parameters
    ----------
    config_file : json
        json file containing the bot_token and the bot_chatID
    """

    def __init__(self, config_file=None) -> None:
        self.__name__ = "telebot"
        if not config_file:
            from .. import TOKENS
            config_file = os.path .join(TOKENS, "telegram.json")
        super().__init__(config_file)
        self._url = self._compose_url()

    def _compose_url(self):
        """compose the base url to send the message

        Returns
        -------
        str
            base url
        """
        return 'https://api.telegram.org/bot' + self._credentials["bot_token"] + '/sendMessage?chat_id=' + self._credentials["bot_chatID"] + '&parse_mode=HTML&text='

    def send_message(self, message):
        """send the message over telegram

        Parameters
        ----------
        bot_message : str
            message to send

        Returns
        -------
        json
            response from the server
        """
        response = requests.get(self._url + message)
        return response.json()


if __name__ == '__main__':

    # init the bot
    bot = TeleBot()

    # send the message
    bot.send_message('ciao mamma')



import os
from git import Repo
from ._bot import BaseBot

COMMITS_TO_PRINT = 5


class DeployBot(BaseBot):
    """Bot to help with the deployment of code on a server. It uses a `sshBot` in
    the background.

    Warnings
    --------
    * You need to have an sshbot settings configured to grant access to the server.
    * You must have :code:git>=2.25 on the server side.

    Parameters
    ----------
    project : str
        project name. It is used to find the configuration file associated to it.
    config_file : json
        path to a json file containing the required info. You can use the command
        line to create one or manually configure a json file following this
        template:

            .. code-block:: json

                {
                    "local_repo_path" :  "_______",
                    "server_repo_path" : "_______",
                    "server_addres" :    "_______",
                }

    Attributes
    ----------
    local_repo_path : str
        path to the local clone of the repository
    server_repo_path : str
        path to the server bare repository.
    server_addres : str
        complete server address (*username@host*).
    server_complete_path : str
        complete path to the server repository (*username@host:pat/to/server/repository*)
    local_repo : obj
        gitpython Repo object pointing to the local repository.

    Examples
    --------
    .. code-block:: python

        bot = DeployBot('my_project')
        bot.deploy_to_server()
    """

    def __init__(self, project=None, config_file=None) -> None:
        self.__name__ = "deploybot"
        if not config_file:
            if not project:
                raise ValueError("Either a project name or a config_file must be passed")
            from .. import TOKENS
            config_file = os.path .join(TOKENS, "{}.json".format(project))
        super().__init__(config_file)
        self.server_complete_path = "{}:{}".format(self.server_address, self.server_repo_path)
        self.local_repo = Repo(self.local_repo_path)

    def deploy_to_server(self, remote_name="deploy", local_name="master", commit_message="deployed"):
        """Deploy the changes to the server.

        Parameters
        ----------
        remote_name : str, optional
            name of the git remote, by default "deploy"
        local_name : str, optional
            name of the local branch to push, by default "master"
        commit_message : str, optional
            message for the commit, by default "deployed"
        """
        if remote_name not in self.local_repo.remotes:
            self.configure_local()
        self._git_push(commit_message, remote_name, local_name)

    def _git_hooks(self):
        raise NotImplementedError()

    def _git_push(self, commit_message, remote_name, local_name, commit=True):
        """Push the local repo to the server

        Parameters
        ----------
        commit_message : str
            message for the commit.
        remote_name : str
            name of the git remote.
        local_name : str, optional
            name of the local branch to push.
        commit : bool, optional
            add untracked files and commit changes, by default True.
        """

        # avoid the case that there is nothing to commit
        if (self.local_repo.index.diff(None) or self.local_repo.untracked_files) and commit:
            self.local_repo.git.add(A=True)
            self.local_repo.git.commit(m=commit_message)
            print('changes committed')
        else:
            print('no changes to commit')
        print('local deployed to server')
        self.local_repo.git.push('--set-upstream', remote_name, local_name)

    # def print_commit(commit):
    #     print('----')
    #     print(str(commit.hexsha))
    #     print("\"{}\" by {} ({})".format(commit.summary,
    #                                      commit.author.name,
    #                                      commit.author.email))
    #     print(str(commit.authored_datetime))
    #     print(str("count: {} and size: {}".format(commit.count(),
    #                                               commit.size)))

    # def print_repository(repo):
    #     print('Repo description: {}'.format(repo.description))
    #     print('Repo active branch is {}'.format(repo.active_branch))
    #     for remote in repo.remotes:
    #         print('Remote named "{}" with URL "{}"'.format(remote, remote.url))
    #     print('Last commit for repo is {}.'.format(str(repo.head.commit.hexsha)))


if __name__ == "__main__":
    bot = DeployBot('my_project')
    bot.deploy_to_server(local_name="main")



"""Console script for codebots.

You can get detaled information about each of the bots with the :code:`--help` option.

.. code-block:: bash

    codebots    --help
    emailbot    --help
    telebot     --help
    slackbot    --help
    sshbot      --help
    deploybot   --help


"""
import sys
import os
from pathlib import Path
import click
import codebots
from codebots.bots import SlackBot
from codebots.bots import TeleBot
from codebots.bots import EmailBot
from codebots.bots import sshBot
from codebots.bots.deploybot import DeployBot
from codebots.utilities.ssh import gen_keypair, add_pubkey_to_server
from codebots.utilities.tokens import add_token, get_telegram_chatid, set_token_dir, reset_token_dir
from codebots.utilities.deploy import configure_local, configure_server


# -------------------------------- MAIN ----------------------------------#
@click.group()
def main():
    """base bot to setup the common settings for all the other bots.

    Run `codebots ono-o-one` for more info.
    """
    pass


@main.command()
def one_o_one():
    """Basic explanation of command line usage."""

    click.echo("\nHey there! this is codebots, a family of bots here to help you!\n\n"
               "To use codebots directly from the command line, type the name of your bot followed by the action.\n"
               "For example, this command sends a message through slack:\n\n"
               "    slackbot --channel=random \"Hello from your slacbot!\"\n")


@main.command()
def get_tokens_path():
    """Get the path to the tokens folder.\n

    Return\n
    ------\n
    str\n
        path to the folder containing the tokens or passwords for the codebots.\n
    """
    out = codebots.TOKENS
    click.echo(out)


@main.command()
@click.argument('path')
def set_tokens_path(path):
    """Set the path to the tokens folder.\n

    Parameters\n
    ----------\n
    path : str\n
        path to the folder containing the tokens or passwords for the codebots.\n
    """
    out = set_token_dir(path)
    click.echo(out)


@main.command()
def reset_tokens_path():
    """Reset the tokens path to the default (~/.tokens)."""
    out = reset_token_dir()
    click.echo(out)

# -------------------------------- SLACK ----------------------------------#


@click.group()
def slackbot():
    """bot to interact with slack"""
    pass


@slackbot.command()
@click.argument('token')
def set_token(token):
    """create the token file with the credentials.\n

    Parameters\n
    ----------\n
    token : str\n
        token of the telegram bot.\n
    chatid : str\n
        chatID of the chat with the bot.\n
    """
    out = add_token("slack", bot_token=token)
    click.echo(out)


@slackbot.command()
@click.option('--channel', default='general', help='the channel you want to send the message to')
@click.argument('message', default='Ciao Mamma!')
def send(message, channel):
    """Send a message using slack.\n
    message : txt\n
        the message you want to send to yo.ur slack channel, by default `Ciao Mamma!`
    """
    bot = SlackBot()
    bot.send_message(channel, message)

# ----------------------------- TELEGRAM ----------------------------------#


@click.group()
def telebot():
    """bot to interact with telegram"""
    pass


@telebot.command()
@click.argument('message', default='Ciao Mamma!')
def send(message):
    """Send a message using slack.\n
    message : txt\n
        the message you want to send to yo.ur slack channel, by default `Ciao Mamma!`
    """

    bot = TeleBot()
    bot.send_message(message)


@telebot.command()
@click.argument('token')
def set_token(token):
    """create the token file with the credentials.\n

    Parameters\n
    ----------\n
    token : str\n
        token of the telegram bot.\n
    chatid : str\n
        chatID of the chat with the bot.\n
    """
    try:
        chatid = get_telegram_chatid(token)
    except Exception:
        raise ConnectionError("Something went wrong! Did you start a chat with your bot on telegram?")
    out = add_token("telegram", bot_token=token, bot_chatID=str(chatid))
    click.echo(out)

# -------------------------------- EMAIL ----------------------------------#


@click.group()
def emailbot():
    """bot to send emails from the command line"""
    pass


@emailbot.command()
@click.argument('receiver', default='mamma@email.com')
@click.argument('subject', default='Ciao')
@click.argument('body', default='Ciao Mamma!')
# @click.option('--attach',  type='path', help='path to any file you want to attach')
def send(receiver, subject, body):
    """Send an email to an email address.\n

        Parameters\n
        ----------\n
        receiver : str\n
            email address of the receiver\n
        subject : str\n
            subject of the email\n
        body : str\n
            body text of the email\n
        attachment : str, optional\n
            path to the file to attach, by default None\
    """

    # sender = Sender.form_file(".tokens/email")
    receiver = "francesco.ranaudo@gmail.com"
    subject = "message from bot"
    body = "This message was sent by a bot"
    # attach = "document.pdf"

    bot = EmailBot()
    bot.send_email(receiver, subject, body)


@emailbot.command()
@click.argument('username')
@click.argument('password')
def set_token(username, password):
    """create the token file with the credentials.\n

    Parameters\n
    ----------\n
    token : str\n
        token of the telegram bot.\n
    chatid : str\n
        chatID of the chat with the bot.\n
    """
    out = add_token("email", username=username, password=password)
    click.echo(out)


# --------------------------------- SSH -----------------------------------#
@click.group()
def sshbot():
    """bot to interact with telegram"""
    pass


@sshbot.command()
@click.argument('hostname')
@click.argument('username')
@click.option('--password', default='', help='password to access the host or to decrypt the private key')
@click.option('--pvtkey', default='', help='path to the private key')
def set_token(hostname, username, password, pvtkey):
    """create the token file with the credentials.\n

    Parameters\n
    ----------\n
    hostname : str\n
        ip address of the server.\n
    username : str\n
        username on the server.\n
    password : str\n
        password on the server, by default empty.\n
    pvtkey : str\n
        path to the private RSA key file, by default empty.\n
    """
    out = add_token("ssh", hostname=hostname, username=username, password=password, pvtkey=pvtkey)
    click.echo(out)


@sshbot.command()
@click.option('--ssh_folder', default=None, help='path where the key pair will be saved, by default None (the `USER/.ssh` folder will be used)')
@click.option('--password', default=None, help='encrypt the private key with a password')
def genkeys(ssh_folder, password):
    """Create a set of public and private keys and save them in the given folder.
    """
    out = gen_keypair(ssh_folder, password)
    click.echo(out)


@sshbot.command()
@click.argument('hostname')
@click.argument('username')
@click.argument('password')
@click.option('--ssh_folder', default=None, help='path where the key pair will be saved')
def link_keys(hostname, username, password, ssh_folder):
    """Adds the public key to the server's list.\n

    Parameters\n
    ----------\n
    hostname : str\n
        ip address of the server.\n
    username : str\n
        username on the server.\n
    password : str\n
        password on the server, by default empty.\n
    """
    bot = sshBot(config_file=None, hostname=hostname, username=username, password=password, pvtkey="")
    if not ssh_folder:
        ssh_folder = os.path.join(str(Path.home()), '.ssh')
    out = add_pubkey_to_server(bot, ssh_folder)
    click.echo(out)
    out = add_token("ssh", hostname=hostname, username=username, password="", pvtkey=os.path.join(ssh_folder, 'id_rsa'))
    click.echo(out)


# ------------------------------- DEPLOY ----------------------------------#
@click.group()
def deploybot():
    """bot to deploy projects to a server"""
    pass


@deploybot.command()
@click.argument('project')
@click.argument('address')
@click.argument('local')
@click.argument('server')
@click.option('--branch', default='main', help='branch to push to.')
@click.option('--sshbot', default=None, help='instance of an `sshBot` with access to the server.')
def configure(project, address, local, server, branch, sshbot):
    """Configure a local repository to sync with a server.\n

    Parameters\n
    ----------\n
    project : str\n
        name of the project for the setting file.\n
    address : str\n
        server address (username@host).\n
    local : str\n
        path to the local clone of the repository.\n
    server : str\n
        path to the server bare repository. If no repository is present\n
        at the given location a bare new one is created.\n
    """
    out = add_token(bot=project, server_address=address, local_repo_path=local, server_repo_path=server)
    click.echo(out)
    bot = DeployBot(project)
    out = configure_local(bot.local_repo, bot.server_complete_path)
    click.echo(out)
    out = configure_server(bot.server_repo_path, branch, sshbot)
    click.echo(out)


@deploybot.command()
@click.argument('project')
def configure_local(project):
    """Configure a local repository to sync with a server.\n

    Parameters\n
    ----------\n
    project : str\n
        name of the project for the setting file.\n
    """
    bot = DeployBot(project)
    out = configure_local(bot.local_repo, bot.server_complete_path)
    click.echo(out)


@deploybot.command()
@click.argument('project')
@click.option('--branch', default='main', help='branch to push to.')
@click.option('--sshbot', default=None, help='instance of an `sshBot` with access to the server.')
def configure_remote(project, branch, sshbot):
    """Configure a local repository to sync with a server.\n

    Parameters\n
    ----------\n
    project : str\n
        name of the project for the setting file.\n
    address : str\n
        server address (username@host).\n
    local : str\n
        path to the local clone of the repository.\n
    server : str\n
        path to the server bare repository. If no repository is present\n
        at the given location a bare new one is created.\n
    """
    bot = DeployBot(project)
    out = configure_server(bot.server_repo_path, branch, sshbot)
    click.echo(out)


# -------------------------------- DEBUG ----------------------------------#
if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import json


class BaseBot():
    """BaseBot class for the other bots.
    """

    def __init__(self, config_file) -> None:
        self._credentials = self._get_token(config_file)
        for k, v in self._credentials.items():
            self.__setattr__(k, v)

    def _get_token(self, config_file):
        """read the access token form a json file.

        Parameters
        ----------
        config_file : json
            json file containing the bot_token and the bot_chatID

        Returns
        -------
        dict
            credential info
        """
        try:
            with open(config_file, "r") as f:
                token = json.load(f)
            return token
        except FileNotFoundError:
            raise FileNotFoundError("there is no token for this bot in the tokens folder.\n\n\
please provide one by running in the command line the following command:\n\
`{} set-token` and provide the required information\n\n\
or set a new tokens folder using:\n\
`codebots set-tokens-path \"path-to-tokens-folder\"`\n".format(self.__name__))



__author__ = """Francesco Ranaudo"""
__email__ = 'francesco.ranaudo@gmail.com'
__version__ = '0.8.3'

import os
import json
from pathlib import Path
from .decorators import *


home = str(Path.home())

HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, '../../'))
DATA = os.path.abspath(os.path.join(HOME, 'data'))
DOCS = os.path.abspath(os.path.join(HOME, 'docs'))
TEMP = os.path.abspath(os.path.join(HOME, 'temp'))
SETTINGS = os.path.join(HERE, './settings.json')


with open(SETTINGS, 'r') as f:
    path = json.load(f)["tokens_path"]

# the default location is in ~/.tokens
if path == "None":
    TOKENS = os.path.join(Path.home(), ".tokens")
else:
    TOKENS = path

if not os.path.isdir(TOKENS):
    Path(TOKENS).mkdir(parents=True, exist_ok=True)



from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
import os


from ._bot import BaseBot

__all__ = [
    'SlackBot'
]


class SlackBot(BaseBot):
    """Bot that automatically sends messages over slack

    Parameters
    ----------
    config_file : file
        file with the access token for the slack workspace.
    """

    def __init__(self, config_file=None) -> None:
        if not config_file:
            from .. import TOKENS
            config_file = os.path.join(TOKENS, "slack.json")
        self.__name__ = "slackbot"
        super().__init__(config_file)
        self._client = self.connect()

    @property
    def client(self):
        """class : slack `WebClient` class"""
        return self._client

    def connect(self):
        """connect to the slack webclient.

        Returns
        -------
        class
            slack WebClient class
        """
        return WebClient(token=self.bot_token)

    def _fetch_channel_id(self, channel, output=False, verbose=False):
        """Retrive the channel ID from its name

        Parameters
        ----------
        channel : str
            slack channel where the message is sent to.
        output : bool, optional
            return the channel id, by default False
        verbose : bool, optional
            print output, by default False

        Returns
        -------
        str
            id of the channel
        """
        result = self.client.conversations_list()
        for chn in result["channels"]:
            if chn["name"] == channel:
                conversation_id = chn["id"]
                # Print result
                if verbose:
                    print(f"Found conversation ID: {conversation_id}")
                if output:
                    return conversation_id
                break
        return conversation_id

    def send_message(self, verbose=False, **kwargs):
        """Send a message to a channel.

        Parameters
        ----------
        message : str, optional
            message to send over slack. If not provided a default message is sent.
        channel : str, optional.
            slack channel where the message is sent to. If not provided, the message
            is sent to `general`.
        verbose : bool, optional
            print WebClient respons, by default False

        Notes
        -----
        Make sure that the bot has been previously added the bot to the channel.
        """
        message = kwargs.get('message', "Ciao! Your results are ready! :)")
        channel = kwargs.get('channel', "general")
        channel_id = self._fetch_channel_id(channel)
        try:
            # Call the conversations.list method using the WebClient
            result = self._client.chat_postMessage(
                channel=channel_id,
                text=message
                # You could also use a blocks[] array to send richer content
            )
            # Print result, which includes information about the message (like TS)
            if verbose:
                print(result)

        except SlackApiError as e:
            print(f"Error: {e}")


# Debug
if __name__ == "__main__":

    bot = SlackBot()
    bot.send_message(channel='topopt', message='test')



import json
import os
import requests
from .. import SETTINGS, TOKENS


def set_token_dir(path):
    """Set the location where codebots looks for the bots credentials.

    Parameters
    ----------
    path : str
        path to the folder where the json files with credentials are stored.

    Returns
    -------
    str
        message
    """
    path_dic = {}
    try:
        with open(SETTINGS, 'w') as f:
            path_dic["tokens_path"] = path
            json.dump(path_dic, f)
        return "tokens path updated!"
    except Exception:
        return "something went wrong!"


def reset_token_dir():
    """Reset the credentials folder to default.

    Returns
    -------
    str
        message
    """
    set_token_dir("None")
    return "tokens path resetted!"


def add_token(bot, **kwargs):
    """Create a json file with the credential settings needed for the bot to
    operate.

    Parameters
    ----------
    bot : str
        name of the bot to configure.

    Examples
    --------
    >>> add_token(telegram, bot_token="some_token", bot_chatID=12345)
    """

    config_dict = {k: v for k, v in kwargs.items()}
    config_file = os.path.join(TOKENS, "{}.json".format(bot))

    try:
        with open(config_file, "w") as f:
            json.dump(config_dict, f)
            return "tokens file for {} updated!".format(bot)
    except Exception:
        return "something went wrong!"


def get_telegram_chatid(token):
    response = requests.get("https://api.telegram.org/bot{}/getUpdates".format(token))
    return response.json()['result'][0]['message']['chat']['id']



import paramiko
import os
import socket


from ._bot import BaseBot

__all__ = [
    'sshBot'
]


class sshBot(BaseBot):
    """sshBot to help with ssh connections to a server.

    Parameters
    ----------
    hostname : str
        ip address of the server, by default None.
    username : str
        username on the server, by default None.
    password : str
        password on the server, by default None.
    pvtkey : str
        path to the private RSA key file, by default None.

    Attributes
    ----------
    hostname : str
        ip address of the server, by default None.
    username : str
        username on the server, by default None.
    password : str
        password on the server, by default None.
    pvtkey : str
        path to the private RSA key file, by default None.
    ssh_client : obj
        paramiko `SSHClient` object, if a connection is active, otherwise None.
    sftp_client : obj
        paramiko `SFTPClient` object, if a connection is active, otherwise None.
    """

    def __init__(self, config_file='default', **kwargs) -> None:

        self.__name__ = "sshbot"
        if not config_file:
            self._credentials = kwargs
            for k, v in self._credentials.items():
                self.__setattr__(k, v)
        elif config_file == 'default':
            from .. import TOKENS
            config_file = os.path .join(TOKENS, "ssh.json")
            super().__init__(config_file)
        else:
            super().__init__(config_file)

        self.host_address = self.hostname if '.' in self.hostname else socket.gethostbyname(self.hostname)
        self._ssh_client = None
        self._sftp_client = None

    # @classmethod
    # def from_credentials_file(cls, json_file):
    #     """Get credentials from json file

    #     Parameters
    #     ----------
    #     json_file : str
    #         path to the json file with the credentials for the server.
    #         it should have the following structure:

    #         .. code-block:: json

    #             {
    #                 "hostname" : "_______",
    #                 "username" : "_______",
    #                 "password" : "_______",
    #                 "pvtkey"   : "_______"
    #             }

    #     Returns
    #     -------
    #     cls
    #         sshBot
    #     """
    #     with open(json_file, 'r') as f:
    #         credentials = json.load(f)
    #     return cls(**credentials)

    @property
    def ssh_client(self):
        return self._ssh_client

    @property
    def sfpt_client(self):
        return self._sfpt_client

    def connect_ssh_client(self):
        """Establish the ssh connection

        Returns
        -------
        obj
            ssh_client obj
        """
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        k = paramiko.RSAKey.from_private_key_file(self.pvtkey, password=self.password) if self.pvtkey else None
        ssh_client.connect(hostname=self.host_address, username=self.username, password=self.password, pkey=k)
        print("\nconnected\n")
        return ssh_client

    def execute_cmds(self, commands, close_connection=True, verbose=True):
        """Execute general command on the server side

        Parameters
        ----------
        commands : list of str
            list of commands (str) to execute on the server.
        close_connection : bool (optional)
            if true close the ssh connection, by default True. Leave the connection
            open if you plan to run several commands.

        Returns
        -------
        None
        """
        if not self._ssh_client:
            self._ssh_client = self.connect_ssh_client()

        out_dict = {"stdout": [], "stderr": []}

        for command in commands:
            if verbose:
                print("Executing {}:\n".format(command))
            stdin, stdout, stderr = self._ssh_client.exec_command(command)
            out_dict["stdout"].append(stdout.read().rstrip().decode("utf-8"))
            out_dict["stderr"].append(stderr.read().rstrip().decode("utf-8"))

            if verbose:
                print(out_dict["stdout"][-1])
                print("\nErrors (if any):")
                print(out_dict["stderr"][-1])

        # close the connection
        if close_connection:
            self._ssh_client.close()
            self._ssh_client = None

        return out_dict

    def connect_sftp_client(self):
        """Connect to the server through SFPT on port 22.

        Returns
        -------
        obj
            paramiko sfpt client object
        """
        transport = paramiko.Transport((self.host_address, 22))
        k = paramiko.RSAKey.from_private_key_file(self.pvtkey, password=self.password) if self.pvtkey else None
        transport.connect(username=self.username, password=self.password, pkey=k)
        return paramiko.SFTPClient.from_transport(transport)

    def get_folder_from_server(self, path, dest, recursive=True, close_connection=True):
        """Retrieve the content of a folder from the server.

        Parameters
        ----------
        sftp_client : obj
            paramiko sftp client object
        path : str
            path to the folder on the server
        dest : str
            path to the folder on the client (local)
        recursive : bool (optional)
            if true get subfolders content, by default
        close_connection : bool (optional)
            if true close the ssh connection, by default True. Leave the connection open if you plan to run several commands.

        Warnings
        --------
        If in any folder there are files without extensions, the code will fail!
        """

        # connect to the server through SFPT
        if not self._sftp_client:
            self._sftp_client = self.connect_sftp_client()
            print("connection enstablished")

        for item in self._sftp_client.listdir(path):
            remotefile = os.path.join(path, str(item))
            if not os.path.isdir(dest):
                os.mkdir(dest)
            localfilepath = os.path.join(dest, str(item))
            # check if it is a folder (note file w/o ext will thorw an exception!)
            if '.' in item:
                self._sftp_client.get(remotefile, localfilepath)
            else:
                if recursive:
                    self.get_folder_from_server(self._sftp_client, remotefile, localfilepath)
        print("files tranferred!")
        # close the connection
        if close_connection:
            self._sftp_client.close()
            self._sftp_client = None
            print("connection closed")


if __name__ == '__main__':
    pass

    # bot = sshBot(config_file=None, hostname="192.168.0.199", username="franaudo", password="Sxcdews23", pvtkey="")
    bot = sshBot()

    bot.execute_cmds(commands=['ls'], close_connection=True)
    # print(bot.ssh_client)
    # bot.execute_cmds(commands=['ls -l'], close_connection=True)
    # print(bot.ssh_client)



import os
from ..bots import sshBot
from pathlib import Path


def configure_local(local_repo, server_complete_path):
    """Configure the local repository to sync with the server.

    Parameters
    ----------
    local_repo_path : str
        path to the local clone of the repository
    server_complete_path : str
        complete path to the server repository (*username@host:pat/to/server/repository*)
    """
    if 'deploy' in local_repo.remotes:
        local_repo.delete_remote("deploy")
    local_repo.create_remote("deploy", server_complete_path)

    return """`deploy` added to remotes.
        local repository configured!"""


def configure_server(server_repo_path, branch='main', sshbot=None, os_type='linux'):
    """Configure the server side:
        - check if server has git repository folder;
        - if not create a bare repository;
        - configure the server repository to sync with the local.

    Parameters
    ----------
    server_repo_path : str
        path to the server bare repository.
    sshbot : obj, optional
        instance of an `sshBot` with access to the server, by default None.
    """

    # git_folder = Path(server_repo_path).joinpath('.git')
    git_folder = server_repo_path + '/.git'
    if not sshbot:
        sshbot = sshBot()

    # check if server folder exists and if it is a git repository
    tests = ["if test -d {}; then echo {}; fi".format(x, "yes") for x in [server_repo_path, git_folder]]
    std_dict = sshbot.execute_cmds(tests, False, False)

    cmds = ['{} {}'.format('mkdir' if os_type == 'windows' else 'mkdir -p', server_repo_path),
            'git --git-dir={} init -b {}'.format(git_folder, branch),
            'git --git-dir={} config receive.denyCurrentBranch updateInstead'.format(git_folder)]

    if std_dict["stdout"][0] != 'yes':
        cmd = cmds
    else:
        cmd = cmds[1:] if std_dict["stdout"][1] != 'yes' else [cmds[-1]]
    out = sshbot.execute_cmds(cmd, verbose=False)

    for x in out["stdout"]:
        if x != '':
            print(x)

    for x in out["stderr"]:
        if x != '':
            return x
    return 'server repository configured!'



from importlib import import_module


def monitor(bot_name, **dec_kwargs):
    def decorator(function):
        def wrapper(*args, **kwargs):
            function(*args, **kwargs)
            cls = getattr(import_module('codebots.bots.{}'.format(bot_name.lower())), bot_name)
            bot = cls()
            if 'message' not in dec_kwargs.keys():
                dec_kwargs['message'] = "{} is done!".format(str(function.__name__))
            bot.send_message(**dec_kwargs)
        return wrapper
    return decorator



import yagmail
import json
import os

from ._bot import BaseBot

__all__ = [
    'Sender',
    'EmailBot'
]


class Sender():
    """Sender class to manage the sender email settings.

    Parameters
    ----------
    username : str
        username used to access the sender email account
    password : str
        password used to access the seder email account
    """

    def __init__(self, username, password) -> None:
        self._username = username
        self._password = password

    @property
    def username(self):
        """str : username used to access the sender email account"""
        return self._username

    @property
    def password(self):
        """str : password used to access the sender email account"""
        return self._password

    @classmethod
    def form_file(cls, config_file):
        """read the access token form a file.

        Returns
        -------
        str
            access token
        """

        with open(config_file, "r") as f:
            credentials = json.load(f)
        user = cls(username=credentials["username"], password=credentials["password"])
        return user


class EmailBot(BaseBot):
    """EmailBot.

    Parameters
    ----------
    sender : obj
        Sender object configured with sender account credentials.
    """

    def __init__(self, config_file=None, sender=None) -> None:
        self.__name__ = "emailbot"
        if not sender:
            if not config_file:
                from .. import TOKENS
                config_file = os.path.join(TOKENS, "email.json")
            super().__init__(config_file)
        else:
            if not isinstance(sender, Sender):
                raise ValueError("the sender is not valid")
            self.sender = sender

    def send_email(self, receiver, subject, body, attachment=None):
        """Send an email to an email address.

        Parameters
        ----------
        receiver : str
            email address of the receiver
        subject : str
            subject of the email
        body : str
            body text of the email
        attachment : str, optional
            path to the file to attach, by default None
        """
        yag = yagmail.SMTP(self.sender.username, self.sender.password)
        yag.send(
            to=receiver,
            subject=subject,
            contents=body,
            attachments=attachment,
        )


if __name__ == "__main__":
    # sender = Sender.form_file("C:/Users/franaudo/.tokens/email.json")
    receiver = "francesco.ranaudo@gmail.com"
    subject = "message from bot"
    body = "This message was sent by a bot"
    filename = "document.pdf"

    # bot = EmailBot(sender)
    bot = EmailBot()
    bot.send_email(receiver, subject, body)
